/* p80211.h
   defs for 802.11 header/frame handling

   Author : KoreK
   Date : Aug 26, 2004
   Licence : GPL v2 or above (see accompanying COPYING file)

   Copyright (c) 2004 KoreK, All Rights reserved
*/


#ifndef P80211_H
#define P80211_H
#include <stdio.h>
#include <sys/types.h>
#include <endian.h>

/* Intended use:
   1) init/build p80211_header, make (alloc) a frame by 
      appending data, inject with wlan, free frame
   2) read a frame from wlan, read relevant information in 
      header/data, free frame
   wlan specific functions are in wlan.c
*/

#if __BYTE_ORDER == __BIG_ENDIAN
#define h_to_le(x) ((((x)>>8) & 0xff) | (((x) << 8) & 0xff00))
#define le_to_h(x) ((((x)>>8) & 0xff) | (((x) << 8) & 0xff00))
#elif __BYTE_ORDER == __LITTLE_ENDIAN
#define h_to_le(x) (x)
#define le_to_h(x) (x)
#else 
#error Undefined endian-ness
#endif



struct p80211_header {
    u_char base[0];
    u_int16_t fc;
    u_int16_t dur_id;
    u_char mac1[6];
    u_char mac2[6];
    u_char mac3[6];
    u_int16_t seq;
    u_char mac4[6];
    u_int16_t data_len;
    u_char null[14];
    u_char data[0];
} __attribute__ ((packed));

typedef struct p80211_header p80211_header_t[1];

typedef struct p80211_header *p80211_frame_t;


/* header manipulation setters */

void p80211_header_init(struct p80211_header *h);

void p80211_set_type(struct p80211_header *h,
		     int type);

void p80211_set_flag(struct p80211_header *h,
		     int flag,
		     int value);

void p80211_set_duration(struct p80211_header *h,
			 int duration);

void p80211_set_fragnb(struct p80211_header *h,
		       int frag_nb);

void p80211_set_seqnb(struct p80211_header *h,
		      int seqnb);

int p80211_set_mac(struct p80211_header *h,
		   int mac_idx,
		   const u_char *MAC);


/* header manipulation getters */

struct p80211_frame_struct p80211_struct(const struct p80211_header *h);

int p80211_type(const struct p80211_header *h);

int p80211_flag(const struct p80211_header *h,
		int flag);


int p80211_duration(const struct p80211_header *h);

int p80211_fragnb(const struct p80211_header *h);

int p80211_seqnb(const struct p80211_header *h);

const u_char *p80211_mac(const struct p80211_header *h,
			 int mac_idx);

int p80211_data_len(const struct p80211_header *h);


/* frame construction / getters / destroy */

p80211_frame_t make_frame(struct p80211_header *h,
			  const u_char *data,
			  size_t data_len);

void free_frame(p80211_frame_t f);



/* dumpers */

void fprint_hex_buffer(FILE *f,
		       const u_char *b,
		       size_t b_len);

void fprint_header(FILE *f,
		   const struct p80211_header *h);


void fprint_frame(FILE *f,
		  const p80211_frame_t frame);


/* constants */

enum fc_flags {
    toDS_f     = h_to_le(1<<8),
    fromDS_f   = h_to_le(1<<9),
    morefrag_f = h_to_le(1<<10),
    retry_f    = h_to_le(1<<11),
    pwrmgt_f   = h_to_le(1<<12),
    moredata_f = h_to_le(1<<13),
    wep_f      = h_to_le(1<<14),
    order_f    = h_to_le(1<<15),
    allflags_f = h_to_le(0xff00)
};


/* enumeration of addressing mode */
enum {
    DA_SA_BSSID_a=0,
    DA_BSSID_SA_a,
    BSSID_SA_DA_a,
    RA_TA_DA_SA_a,
    BSSID_TA_a,
    RA_BSSID_a,
    RA_TA_a,
    RA_a
};


enum {
    DA=0,SA,BSSID,RA,TA
};



/* structure to interpret a 802.11 frame : 
   Given [type_subtype][DSto/from] returns:
   1) type of the frame
   2) length of the header
   3) tag interpretation for mac1,mac2,mac3,mac4
   4) 
*/
struct p80211_frame_struct {
    unsigned int type:5;

    unsigned int header_len:5;        //header_len is fuction of mac_order 
    unsigned int mac_order:3;

    unsigned int body:1;
    unsigned int min_body_len:5;
    unsigned int max_body_len:12;
} __attribute__ ((packed));



/* enumration of frame types */
enum {
    INVALID_FRAME = 0,
    ASSOC_REQ_FRAME,
    DATA_FRAME,
    ASSOC_RESP_FRAME,
    DATA_CFACK_FRAME,
    REASSOC_REQ_FRAME,
    DATA_POLL_FRAME,
    REASSOC_RESP_FRAME,
    DATA_CFACK_POLL_FRAME,
    PROBE_REQ_FRAME,
    NULL_FRAME,
    PROBE_RESP_FRAME,
    NULL_CFACK_FRAME,
    NULL_POLL_FRAME,
    NULL_CFACK_POLL_FRAME,
    BEACON_FRAME,
    ATIM_FRAME,
    DISASSOC_FRAME,
    PS_POLL_FRAME,
    AUTH_FRAME,
    RTS_FRAME,
    DEAUTH_FRAME,
    CTS_FRAME,
    ACK_FRAME,
    CFEND_FRAME,
    CFEND_ACK_FRAME 
};


#endif // P80211_H


