/* pcap_readwrite.c
   reads/writes a pcap file into/from a circular list of packets. 

   Author : KoreK
   Date : Aug 22, 2004
   Licence : GPL v2 or above (see accompanying COPYING file)

   Copyright (c) 2004 KoreK, All Rights Reserved

   Changelog:
   Aug 22, 2004: Initial version
*/

/* Note: 
   ifdefs for prism/wlan packet are necessary when compiling against old
   pcap. Too much caution. Decide before RELEASE. Probably use an 
   up-to-date pcap, or what? 
   2) For the moment, code is endian incompatible. pcap file from 
   a little-endian won't work on a big-endian, and vice-versa.
   Solution: write up a pcap endian inverter? 
   
*/
#define PCAP_READWRITE_C

#include <netinet/in.h>
#include <stdlib.h>
#include <string.h>
#include <pcap.h>
#include "pcap_readwrite.h"


/* given a filename, and an initialized (struct pcap_capture) list 
   append to the list the libpcap packets in the the file 
   return 0 on failure */
int read_pcapfile(const char *filename,
		  struct pcap_capture *capture)
{
    
    pcap_t *W;
    char errbuf[PCAP_ERRBUF_SIZE];
    void (*packet_parse)(u_char *,const struct pcap_pkthdr *,const u_char *)=0;

    W=pcap_open_offline(filename,errbuf);
    if (W==0) {
        fprintf(stderr,"Error open_offline: %s\n",errbuf);
        return -1;
    }

    if (pcap_is_swapped(W)) return -2;

    capture->cap_len=pcap_snapshot(W);
    capture->linktype=pcap_datalink(W);

    switch (pcap_datalink(W)) {
	// Monitor mode - simple 802.11
	case DLT_IEEE802_11:
	    packet_parse=packet_parse_80211;
	    break;
	// Promiscuous associated wlan
	case DLT_EN10MB:
	    packet_parse=packet_parse_8023;
	    break;
	default: 
	    packet_parse=0;
    }

    if (packet_parse) pcap_dispatch(W,-1,packet_parse,(u_char *)capture);

    pcap_close(W);
    return !packet_parse;
}


void packet_parse_80211(u_char *pcap_capture_list,
			const struct pcap_pkthdr *hdr,
			const u_char *data)
{
    struct pcap_capture *capture=(struct pcap_capture *)pcap_capture_list;
    struct pcap_node *node;
    struct packet *p;
    struct p80211_header w_hdr[1];

    capture->pkt++;

    node=alloc_pcap_node(hdr,data);
    // no way to stop the pcap_dispatch, just ignore 
    if (node == 0) return;
    
    capture->pkt_read++;
    append_pcap_node_capture(node,capture);

    // discard truncated pcap packets
    if (hdr->caplen < hdr->len) {
	memset(node->w_packet,0,sizeof *(node->w_packet));
	return;
    }

    grab_80211_header(data,hdr->caplen,w_hdr);

    if ((w_hdr->type == p80211_TYPE_DATA) &&
	(w_hdr->fragmented == 0) &&
	(w_hdr->wep == 1)) {
	
	p=node->w_packet;

	p->frame_len = hdr->caplen - w_hdr->header_len;
	p->frame=node->pkt_body+w_hdr->header_len;
	p->iv=0;
	memcpy(p->src_mac,w_hdr->src_mac,6);
	memcpy(p->dst_mac,w_hdr->dst_mac,6);
	if (w_hdr->bssid_mac) memcpy(p->bssid_mac,w_hdr->bssid_mac,6);
	else memset(p->bssid_mac,0,6);
    }
    else {
	memset(node->w_packet,0,sizeof *(node->w_packet));
    }
    return;
}

/* Format of 802.3 promiscuous mode : 
   1) 802.3 header
      - 6-byte dest MAC
      - 6-byte src MAC
      - 2-/byte network order 802.3 frame length
   2) 802.3 frame
      - IV + key index (4 bytes)
      - wireless encrypted frame
*/
void packet_parse_8023(u_char *pcap_capture_list,
		       const struct pcap_pkthdr *hdr,
		       const u_char *data)
{
    struct pcap_capture *capture=(struct pcap_capture *)pcap_capture_list;
    struct pcap_node *node;

    capture->pkt++;

    node=alloc_pcap_node(hdr,data);
    // no way to stop the pcap_dispatch, just ignore 
    if (node == 0) return;

    capture->pkt_read++;
    append_pcap_node_capture(node,capture);

    // discard truncated pcap packets
    if (hdr->caplen < hdr->len) {
	memset(node->w_packet,0,sizeof *(node->w_packet));
	return;
    } 
    
    // check for cap_len / frame_len consistency
  
    node->w_packet->frame=node->pkt_body+14;
    node->w_packet->frame_len=(data[12] << 8) + data[13];
    node->w_packet->iv=0;
    memcpy(node->w_packet->src_mac,node->pkt_body+6,6);
    memcpy(node->w_packet->dst_mac,node->pkt_body,6);
    memset(node->w_packet->bssid_mac,0,6);
    
    return;
}   


void grab_80211_header(const u_char *data,
		       size_t data_len,
		       struct p80211_header *w_hdr)
{
    int version,type,subtype;
    int tofromDS,more_frag,retry,pwr_mgt,more_data,wep,order;
    size_t header_len;

    // first offset: dst_mac, second: src_mac
    const static size_t dst_mac[4]={4,4,16,16};
    const static size_t src_mac[4]={10,16,10,24};
    const static size_t bssid_mac[4]={16,10,4,0};


    memset(w_hdr,0,sizeof *w_hdr);

    if (data_len<10) {
	w_hdr->type=p80211_TYPE_INVALID;
	return;
    }

    version=data[0] & 0x03;
    type   =(data[0] >> 2) & 0x03;
    subtype=(data[0] >> 4) & 0x0f;

    order    =(data[1] >> 7) & 0x01;
    wep      =(data[1] >> 6) & 0x01;
    more_data=(data[1] >> 5) & 0x01;
    pwr_mgt  =(data[1] >> 4) & 0x01;
    retry    =(data[1] >> 3) & 0x01;
    more_frag=(data[1] >> 2) & 0x01;
    tofromDS =(data[1] ) & 0x03;

    if (version != 0) {
	w_hdr->type=p80211_TYPE_INVALID;
	return;
    }

    w_hdr->type=type;
    if (type != p80211_TYPE_DATA) return;

    if (subtype & 0x08) {
	w_hdr->type=p80211_TYPE_INVALID;
	return;
    }

    header_len= (tofromDS != 3) ? 24 : 30;

    if (data_len<header_len) {
	w_hdr->type=p80211_TYPE_INVALID;
	return;
    }

    // skip null data
    if (subtype & 0x04)    {
	w_hdr->type=p80211_TYPE_SKIP;
	return;
    }

    w_hdr->header_len=header_len;
    w_hdr->fragmented=more_frag;
    w_hdr->wep=wep;
    w_hdr->src_mac=data+src_mac[tofromDS];
    w_hdr->dst_mac=data+dst_mac[tofromDS];
    w_hdr->bssid_mac= (tofromDS != 3) ? data+bssid_mac[tofromDS] : 0;

    return;
}


int write_pcapfile(const char *filename,struct pcap_capture *capture)
{
    pcap_t *W;
    pcap_dumper_t *W_dumper;
    char errbuf[PCAP_ERRBUF_SIZE];
    struct pcap_node *n;

    W=pcap_open_dead(capture->linktype,capture->cap_len);
    // I wish W->errbuf=errbuf; would work, but too much trouble. 

    W_dumper=pcap_dump_open(W,filename);
    if (W_dumper == 0) {
        fprintf(stderr,"Error pcap_dump_open\n");
	pcap_close(W);
        return 0;
    }

    if (W_dumper == 0) {
        fprintf(stderr,"Error open_offline: %s\n",errbuf);
	pcap_close(W);
        return 0;
    }

    for (n=capture->first; n!=0; n=n->next) 
	pcap_dump((u_char *)W_dumper,n->pkt_header,n->pkt_body);

    pcap_dump_close(W_dumper);
    pcap_close(W);
    return 1;
}





/* struct pcap_node are allocated in one chunk 
   containing the pointers and the contents - Anyway 
   they should be created/freed only with the following 
   functions */


void init_pcap_capture(struct pcap_capture *p)
{
    memset(p,0,sizeof *p);
}


struct pcap_node *alloc_pcap_node(const struct pcap_pkthdr *pcap_header,
				  const u_char *pcap_body)
{
    void *base;
    struct pcap_node *pn;
    size_t size;

    if ((pcap_header == 0) || (pcap_body == 0)) return 0;

    size=ALIGN(sizeof(struct pcap_node),DEFAULT_ALIGN) +
	ALIGN(sizeof(struct packet),DEFAULT_ALIGN) +
	ALIGN(sizeof(struct pcap_pkthdr),DEFAULT_ALIGN) +
	pcap_header->caplen;

    base=malloc(size);
    if (base == 0) return 0;

    pn=(struct pcap_node *)base;
    
    base=base+ALIGN(sizeof(struct pcap_node),DEFAULT_ALIGN);
    pn->w_packet=(struct packet *)base;

    base=base+ALIGN(sizeof(struct packet),DEFAULT_ALIGN);
    pn->pkt_header=(struct pcap_pkthdr *)base;
    memcpy(pn->pkt_header,pcap_header,sizeof *pcap_header);


    base=base+ALIGN(sizeof(struct pcap_pkthdr),DEFAULT_ALIGN);
    pn->pkt_body=(u_char *)base;
    memcpy(pn->pkt_body,pcap_body,pcap_header->caplen);
   
    return pn;
}


void free_pcap_node(struct pcap_node *p)
{
    free(p);
}


void free_pcap_capture(struct pcap_capture *p)
{
    struct pcap_node *c,*n;

    c=p->first;
    while (c != 0) {
	n=c->next;
	free_pcap_node(c);
	c=n;
    }
    p->first=0;
    p->last=0;
    p->pkt=0;
    p->pkt_read=0;
}


int append_pcap_node_capture(struct pcap_node *packet,
			     struct pcap_capture *capture)
{
    if (capture->first == 0) {
	if (capture->last == 0) {
	    capture->first=packet;
	    capture->last=packet;
	    packet->previous=0;
	    packet->next=0;
	    return 0;
	}
	else return 1;
    }
    else {
	struct pcap_node *last;
	last=capture->last;
	last->next=packet;
	packet->previous=last;
	packet->next=0;
	capture->last=packet;
	return 0;
    }
}



	
	
